package org.netkernel.lang.trl.endpoint;

import java.io.ByteArrayInputStream;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.w3c.dom.Document;

public class TRLRuntime extends StandardAccessorImpl
{
	Pattern mRequestPattern=Pattern.compile("(\\$e?\\{)(.*?)\\}",Pattern.DOTALL|Pattern.MULTILINE);
	Pattern mDeclarativeRequestPattern=Pattern.compile(".*?<request>.*?",Pattern.DOTALL|Pattern.MULTILINE);
	
	public TRLRuntime()
	{	this.declareThreadSafe();		
	}

	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	String t=context.source("arg:template", String.class);
		INKFResponse resp=context.createResponseFrom(recurse(t, context, context.getThisRequest().argumentExists("tolerant")));
		String mimetype = context.getThisRequest().getArgumentValue("mimetype");
		if (mimetype!=null)
		{	resp.setMimeType(mimetype);
		}
		else
		{	resp.setMimeType("text/plain");
		}
	}
	
	private String recurse(String source, INKFRequestContext aContext, boolean tolerant) throws Exception
	{	source=stripComments(source);
		Matcher m=mRequestPattern.matcher(source);
		int start=0;
		while(m.find(start))
		{	String r=m.group(2);
			String head=m.group(1);
			String result=sourceRequest(r, aContext, tolerant);
			String sub=recurse(result, aContext, tolerant);
			if(head.equals("$e{"))
			{	sub=HTMLEncode(sub);				
			}
			source=m.replaceFirst(sub);
			start=m.start()+sub.length();
			m.reset(source);
		}
		return source;
	}

	private String sourceRequest(String r, INKFRequestContext aContext, boolean tolerant) throws Exception
	{	Matcher m=mDeclarativeRequestPattern.matcher(r);
		String rep=null;
		try
		{
			if(m.matches())
			{	Document d=XMLUtils.parse(new ByteArrayInputStream(r.getBytes()));
				RequestBuilder b=new RequestBuilder(d, aContext.getKernelContext().getKernel().getLogger());
				INKFRequest req=b.buildRequest(aContext, null, null);
				req.setRepresentationClass(String.class);
				rep=(String)aContext.issueRequest(req);
			}
			else
			{	rep=aContext.source(r,String.class);
			}
		}
		catch(NKFException e)
		{	if(tolerant)
			{	rep="{TLR-ERROR}"+e.getDeepestMessage()+"{/TLR-ERROR}";
			}
			else
			{	throw e;
			}
		}
		return rep;
	}
	
	private String stripComments(String source) throws Exception
	{	StringReader sr=new StringReader(source);
		LineNumberReader lnr=new LineNumberReader(sr);
		StringBuilder b=new StringBuilder();
		String line=lnr.readLine();
		String nextline=null;
		while(line!=null)
		{	nextline=lnr.readLine();
			if(!line.startsWith("#"))
			{	b.append(line);
				if(nextline!=null)
				{	b.append("\n");
				}
			}
			line=nextline;
		}
		return b.toString();
	}
	
	public static String HTMLEncode(String s)
	{   if(s==null)
		{	return null;
		}
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < s.length(); i++)
        {   char c = s.charAt(i);
            if (c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z' || c >= '0' && c <= '9')
            {	sb.append(c);
            }
            else
            {	sb.append("&#"+(int)c+";");
            }
        }
        return sb.toString();
    }
}
